//////////////////////////////////////////////////////////////////////////////////
// Dynamic Audio Normalizer - Java/JNI Wrapper
// Copyright (c) 2015 LoRd_MuldeR <mulder2@gmx.de>. Some rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// http://opensource.org/licenses/MIT
//////////////////////////////////////////////////////////////////////////////////

package com.muldersoft.dynaudnorm.test;

import static org.junit.Assert.fail;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

import org.junit.After;
import org.junit.Before;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;

import com.muldersoft.dynaudnorm.JDynamicAudioNormalizer;
import com.muldersoft.dynaudnorm.JDynamicAudioNormalizer.Logger;

@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class JDynamicAudioNormalizerTest
{
	//------------------------------------------------------------------------------------------------
	// PCM File Reader
	//------------------------------------------------------------------------------------------------

	static class PCMFileReader
	{
		private static final int FRAME_SIZE = 4;
		private final BufferedInputStream inputStream;
		private byte [] tempBuffer = null;
		
		public PCMFileReader(final String fileName) throws IOException
		{
			inputStream = new BufferedInputStream(new FileInputStream(new File(fileName)));
		}
		
		public int read(double [][] buffer) throws IOException
		{
			if(buffer.length != 2)
			{
				throw new RuntimeException("Output array dimension must be two!");
			}
			
			final int maxReadSize = buffer[0].length * FRAME_SIZE;
			if((tempBuffer == null) || (tempBuffer.length < maxReadSize))
			{
				tempBuffer = new byte[maxReadSize];
			}
			
			final int sampleCount = inputStream.read(tempBuffer, 0, maxReadSize) / FRAME_SIZE;
			
			if(sampleCount > 0)
			{
				ByteBuffer byteBuffer = ByteBuffer.wrap(tempBuffer);
				byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
				for(int i = 0; i < sampleCount; i++)
				{
					for(int c = 0; c < 2; c++)
					{
						buffer[c][i] = ShortToDouble(byteBuffer.getShort());
					}
				}
			}
			
			return sampleCount;
		}
		
		public void close() throws IOException
		{
			inputStream.close();
		}
		
		private static double ShortToDouble(final short x)
		{
			return ((double)x) / ((double)Short.MAX_VALUE);
		}
	}
	
	//------------------------------------------------------------------------------------------------
	// PCM File Writer
	//------------------------------------------------------------------------------------------------

	static class PCMFileWriter
	{
		private static final int FRAME_SIZE = 4;
		private final BufferedOutputStream outputStream;
		private byte [] tempBuffer = null;
		
		public PCMFileWriter(final String fileName) throws IOException
		{
			outputStream = new BufferedOutputStream(new FileOutputStream(new File(fileName)));
		}
		
		public void write(double [][] buffer, final int sampleCount) throws IOException
		{
			if(buffer.length != 2)
			{
				throw new IOException("Input array dimension must be two!");
			}
			
			final int writeSize = sampleCount * FRAME_SIZE;
			if((tempBuffer == null) || (tempBuffer.length < writeSize))
			{
				tempBuffer = new byte[writeSize];
			}
			
			if(sampleCount > 0)
			{
				ByteBuffer byteBuffer = ByteBuffer.wrap(tempBuffer);
				byteBuffer.order(ByteOrder.LITTLE_ENDIAN);
				for(int i = 0; i < sampleCount; i++)
				{
					for(int c = 0; c < 2; c++)
					{
						byteBuffer.putShort(doubleToShort(buffer[c][i]));
					}
				}
			}
			
			outputStream.write(tempBuffer, 0, writeSize);
		}
		
		public void flush() throws IOException
		{
			outputStream.flush();
		}
		
		public void close() throws IOException
		{
			flush();
			outputStream.close();
		}

		private static short doubleToShort(final double x)
		{
			return (short) Math.round(Math.max(-1.0,  Math.min(1.0, x)) * ((double)Short.MAX_VALUE));
		}
	}
	
	//------------------------------------------------------------------------------------------------
	// Test Functions
	//------------------------------------------------------------------------------------------------

	@Before
	public void setUp()
	{
		System.out.println("\n============================= TEST BEGIN =============================");
	}
	
	@After
	public void shutDown()
	{
		System.out.println("============================= TEST DONE! =============================\n");
	}

	@Test
	public void test1_setLoggingHandler()
	{
		JDynamicAudioNormalizer.setLoggingHandler
		(
			new Logger()
			{
				public void log(final int level, final String message)
				{
					String label = "???";
					switch(level)
					{
						case 0: label = "DBG"; break;
						case 1: label = "WRN"; break;
						case 2: label = "ERR"; break;
					}
					System.err.printf("[JDynAudNorm][%s] %s\n", label, message);
				}
			}
		);
		System.out.println("Message handler installed successfully.");
	}

	@Test
	public void test2_getVersionInfo()
	{
		final int[] versionInfo = JDynamicAudioNormalizer.getVersionInfo();
		System.out.printf("Library Version: %d.%02d-%d\n", versionInfo[0], versionInfo[1], versionInfo[2]);
	}
	
	@Test
	public void test3_getBuildInfo()
	{
		final Map<String,String> buildInfo = JDynamicAudioNormalizer.getBuildInfo();
		for(final String key : buildInfo.keySet())
		{
			System.out.println("Build Info: " + key + "=" + buildInfo.get(key));
		}
	}
	
	@Test
	public void test4_getConfiguration()
	{
		JDynamicAudioNormalizer instance = new JDynamicAudioNormalizer(2, 44100, 500, 31, 0.95, 10.0, 0.0, 0.0, true, false, false);
		final Map<String,Integer> buildInfo = instance.getConfiguration();
		for(final String key : buildInfo.keySet())
		{
			System.out.printf("Configuration: %s=%s\n", key, buildInfo.get(key));
		}
		instance.close();
	}
	
	@Test
	public void test5_getInternalDelay()
	{
		JDynamicAudioNormalizer instance = new JDynamicAudioNormalizer(2, 44100, 500, 31, 0.95, 10.0, 0.0, 0.0, true, false, false);
		final long delayedSamples = instance.getInternalDelay();
		System.out.printf("Delayed Samples: %d\n", delayedSamples);
		instance.close();
	}
	
	@Test
	@SuppressWarnings("deprecation")
	public void test6_constructorAndRelease()
	{
		final int MAX_ROUNDS = 512;
		final int MAX_INSTANCES = 56;
		
		for(int i = 0; i < MAX_ROUNDS; i++)
		{
			System.out.println("--------------------------------------------------------------------------");
			System.out.printf("Round %d of %d\n", i+1, MAX_ROUNDS);
			System.out.println("--------------------------------------------------------------------------");

			Queue<JDynamicAudioNormalizer> instances = new LinkedList<JDynamicAudioNormalizer>();

			// Constructor
			for(int k = 0; k < MAX_INSTANCES; k++)
			{
				JDynamicAudioNormalizer instance = new JDynamicAudioNormalizer(2, 44100, 500, 31, 0.95, 10.0, 0.0, 0.0, true, false, false);
				System.out.printf("Handle: %08X\n", instance.getHandle());
				instances.add(instance);
			}
			
			// Release
			while(!instances.isEmpty())
			{
				instances.poll().close();
			}
		}
	}
	
	@Test
	public void test7_processAudioFile()
	{
		//Open input and output files
		System.out.println("Opening input and out files...");
		PCMFileReader reader = null;
		try
		{
			reader = new PCMFileReader("Input.pcm");
		}
		catch (Exception e)
		{
			fail("Failed to open ionput audio file!");
		}
		
		PCMFileWriter writer = null;
		try
		{
			writer = new PCMFileWriter("Output.pcm");
		}
		catch (Exception e)
		{
			fail("Failed to open ionput audio file!");
		}
		
		double [][] sampleBuffer = new double[2][4096];
		JDynamicAudioNormalizer instance = new JDynamicAudioNormalizer(2, 44100, 500, 31, 0.95, 10.0, 0.0, 0.0, true, false, false);
		
		//Process samples
		System.out.println("Processing input samples...");
		for(;;)
		{
			int sampleCount = 0;
			try
			{
				sampleCount = reader.read(sampleBuffer);
			}
			catch(IOException e)
			{
				fail("Failed to read from input file!");
			}
			
			if(sampleCount > 0)
			{
				final long outputSamples = instance.processInplace(sampleBuffer, sampleCount);
				if(outputSamples > 0)
				{
					try
					{
						writer.write(sampleBuffer, (int) outputSamples);
					}
					catch (IOException e)
					{
						fail("Failed to write to output file!");
					}
				}
			}

			if(sampleCount < sampleBuffer[0].length)
			{
				break; /*EOF reached*/
			}
		}
		
		//Flush pending samples
		System.out.println("Flushing remaining samples...");
		for(;;)
		{
			final long outputSamples = instance.flushBuffer(sampleBuffer);
			if(outputSamples > 0)
			{
				try
				{
					writer.write(sampleBuffer, (int) outputSamples);
				}
				catch (IOException e)
				{
					fail("Failed to write to output file!");
				}
			}
			else
			{
				break; /*No more samples*/
			}
		}
		
		//Reset
		System.out.println("Resetting normalizer instance...");
		instance.reset(); /*This is NOT normally required, but we do it here to test the reset() API*/

		//Close input and output
		System.out.println("Shutting down...");
		try
		{
			reader.close();
			writer.close();
		}
		catch(IOException e)
		{
			fail("Failed to close/flush the files!");
		}
		
		//Finished
		instance.close();
		System.out.println("Completed.");
	}
}
